--
-- Copyright (c) 2009-2010, ERICSSON AB All rights reserved.
-- 
-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
-- 
--     * Redistributions of source code must retain the above copyright notice,
--       this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the ERICSSON AB nor the names of its contributors
--       may be used to endorse or promote products derived from this software
--       without specific prior written permission.
-- 
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-- BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
-- OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-- THE POSSIBILITY OF SUCH DAMAGE.
--

{-# LANGUAGE TypeFamilies #-}

module Feldspar.Compiler.Plugins.PrettyPrint where

import Feldspar.Compiler.PluginArchitecture
import Feldspar.Compiler.Options

-- ===========================================================================
--  == PrettyPrint plugin
-- ===========================================================================

instance Default IsRestrict where
    defaultValue = NoRestrict


instance Default IsDefaultArraySize where
    defaultValue = NoDefaultArraySize


data PrettyPrint = PrettyPrint


instance TransformationPhase PrettyPrint where
    type From PrettyPrint = ()
    type To PrettyPrint = PrettyPrintSemanticInfo
    type Downwards PrettyPrint = (IsRestrict, Int)
    type Upwards PrettyPrint = ()
    
    transformFormalParameter _ (platform,defArrSize) _ up =
        FormalParameter {
            formalParameterVariable =  addDefaultArraySizes v defArrSize,
            formalParameterSemInf = platform 
        }
      where
        v = recursivelyTransformedFormalParameterVariable up
    
    transformLocalDeclaration _ (_,defArrSize) _ up =
        LocalDeclaration {
            localVariable = addDefaultArraySizes v defArrSize,
            localInitValue = recursivelyTransformedLocalInitValue up,
            localDeclarationSemInf = () 
        }
      where
        v = recursivelyTransformedLocalVariable up


instance Plugin PrettyPrint where
    type ExternalInfo PrettyPrint = (Platform,Int)
    executePlugin PrettyPrint (platform,defArrSize) procedure = fst
        $ executeTransformationPhase PrettyPrint (isRestrict platform,defArrSize) procedure where


addDefaultArraySizes :: (SemanticInfo t) => Variable t -> Int -> Variable t
addDefaultArraySizes v defArrSize = v{variableType = addDefaultArraySizes' t}
  where
    t = variableType v
    addDefaultArraySizes' (ImpArrayType (Norm n) t) = ImpArrayType (Norm n) $ addDefaultArraySizes' t
    addDefaultArraySizes' (ImpArrayType Undefined t)  = ImpArrayType (Defined defArrSize) $ addDefaultArraySizes' t
    addDefaultArraySizes' t                         = t

